"""
Tools for exporting PySB models to a variety of other formats.

Exporting can be performed at the command-line or programmatically/interactively
from within Python.

Command-line usage
==================

At the command-line, run as follows::

    python -m pysb.export model.py <format>

where ``model.py`` is a file containing a PySB model definition (i.e.,
contains an instance of ``pysb.core.Model`` instantiated as a global variable).
``[format]`` should be the name of one of the supported formats:

- ``bngl``
- ``bng_net``
- ``kappa``
- ``potterswheel``
- ``sbml``
- ``python``
- ``mathematica``
- ``matlab``

In all cases, the exported model code will be printed to standard
out, allowing it to be inspected or redirected to another file.

Interactive usage
=================

Export functionality is implemented by this module's top-level function
``export``. For example, to export the "Robertson" example model as SBML, first
import the model::

    from pysb.examples.robertson import model

Then import the ``export`` function from this module::

    from pysb.export import export

Call the ``export`` function, passing the model instance and a string
indicating the desired format, which should be one of the ones indicated
in the list in the "Command-line usage" section above::

    sbml_output = export(model, 'sbml')

The output (a string) can be inspected or written to a file, e.g. as follows::

    with open('robertson.sbml', 'w') as f:
        f.write(sbml_output)

Implementation of specific exporters
====================================

Information on the implementation of specific exporters can be found in the
documentation for the exporter classes in the package :py:mod:`pysb.export`:

.. toctree::
   :maxdepth: 2

   sbml
   matlab
   mathematica
   potterswheel
   bngl
   bng_net
   kappa
   python
"""

import re
import textwrap

class Exporter(object):
    """Base class for all PySB model exporters.

    Export functionality is implemented by subclasses of this class. The
    pattern for model export is the same for all exporter subclasses: a
    model is passed to the exporter constructor and the ``export`` method
    on the instance is called.

    Parameters
    ----------
    model : pysb.core.Model
        The model to export.
    docstring : string (optional)
        The header comment to include at the top of the exported file.

    Examples
    --------

    Exporting the "Robertson" example model to SBML using the ``SbmlExporter``
    subclass::

    >>> from pysb.examples.robertson import model
    >>> from pysb.export.sbml import SbmlExporter
    >>> e = SbmlExporter(model)
    >>> sbml_output = e.export()
    """

    def __init__(self, model, docstring=None):
        self.model = model
        """The model to export."""
        self.docstring = docstring
        """Header comment to include at the top of the exported file."""

    def export(self):
        """The export method, which must be implemented by any subclass.

        All implementations of this method are expected to return a single
        string containing the representation of the model in the desired
        format.
        """
        raise NotImplementedError()

# Define a dict listing supported formats and the names of the classes
# implementing their export procedures
formats = {
        'bngl': 'BnglExporter',
        'bng_net': 'BngNetExporter',
        'kappa': 'KappaExporter',
        'potterswheel': 'PottersWheelExporter',
        'sbml': 'SbmlExporter',
        'python': 'PythonExporter',
        'mathematica': 'MathematicaExporter',
        'matlab': 'MatlabExporter',
        }

def export(model, format, docstring=None):
    """Top-level function for exporting a model to a given format.

    Parameters
    ----------
    model : pysb.core.Model
        The model to export.
    format : string
        A string indicating the desired export format.
    docstring : string (optional)
        The header comment to include at the top of the exported file.
    """

    # Import the exporter module. This is done at export runtime to avoid
    # circular imports at module loading
    export_module = __import__('pysb.export.' + format,
                               fromlist=[formats[format]])
    export_class = getattr(export_module, formats[format])
    e = export_class(model, docstring)
    return e.export()

def pad(text, depth=0):
    "Dedent multi-line string and pad with spaces."
    text = textwrap.dedent(text)
    text = re.sub(r'^(?m)', ' ' * depth, text)
    text += '\n'
    return text



from pysb.testing import *
from pysb import *
from pysb.bng import *
from pysb.core import as_complex_pattern

@with_model
def test_generate_network():
    Monomer('A')
    assert_raises(NoInitialConditionsError, generate_network, model)
    Parameter('A_0', 1)
    Initial(A(), A_0)
    assert_raises(NoRulesError, generate_network, model)
    Parameter('k', 1)
    Rule('degrade', A() >> None, k)
    ok_(generate_network(model))

@with_model
def test_compartment_species_equivalence():
    Parameter('p', 1)
    Monomer('Q', ['x'])
    Monomer('R', ['y'])
    Compartment('C', None, 3, p)
    Rule('bind', Q(x=None) + R(y=None) >> Q(x=1) % R(y=1), p)
    Initial(Q(x=None) ** C, p)
    Initial(R(y=None) ** C, p)
    generate_equations(model)
    for i, (cp, param) in enumerate(model.initial_conditions):
        ok_(cp.is_equivalent_to(model.species[i]))
    ok_(model.species[2].is_equivalent_to(Q(x=1) ** C % R(y=1) ** C))



"""Simple model demonstrating expressions that reference observables."""

from pysb import *

Model()

Monomer('Bax', ['conf'], {'conf': ['c0', 'c1', 'c2']})

Initial(Bax(conf='c0'), Parameter('Bax_0', 1))

Parameter('k1', 1)
Parameter('k2', 1)
Parameter('c0_scaling', 0)
Parameter('c1_scaling', 2)
Parameter('c2_scaling', 5)

Rule('c0_to_c1', Bax(conf='c0') >> Bax(conf='c1'), k1)
Rule('c1_to_c2', Bax(conf='c1') >> Bax(conf='c2'), k2)

Observable('Bax_c0', Bax(conf='c0'))
Observable('Bax_c1', Bax(conf='c1'))
Observable('Bax_c2', Bax(conf='c2'))

Expression('NBD_signal',
           Bax_c0 * c0_scaling + Bax_c1 * c1_scaling + Bax_c2 * c2_scaling)



import pysb.core
import pysb.bng
import numpy
from scipy.integrate import ode
from scipy.weave import inline
import scipy.weave.build_tools
import distutils.errors
import sympy
import re
import itertools

# some sane default options for a few well-known integrators
default_integrator_options = {
    'vode': {
        'method': 'bdf',
        'with_jacobian': True,
        # Set nsteps as high as possible to give our users flexibility in
        # choosing their time step. (Let's be safe and assume vode was compiled
        # with 32-bit ints. What would actually happen if it was and we passed
        # 2**64-1 though?)
        'nsteps': 2**31 - 1,
        },
    'cvode': {
        'method': 'bdf',
        'iteration': 'newton',
        },
    }

class Solver(object):
    """An interface for numeric integration of models.

    Parameters
    ----------
    model : pysb.Model
        Model to integrate.
    tspan : vector-like
        Time values over which to integrate. The first and last values define
        the time range, and the returned trajectories will be sampled at every
        value.
    integrator : string, optional
        Name of the integrator to use, taken from the list of integrators known
        to :py:class:`scipy.integrate.ode`.
    integrator_params
        Additional parameters for the integrator.

    Attributes
    ----------
    model : pysb.Model
        Model passed to the constructor
    tspan : vector-like
        Time values passed to the constructor.
    y : numpy.ndarray
        Species trajectories. Dimensionality is ``(len(tspan),
        len(model.species))``.
    yobs : numpy.ndarray with record-style data-type
        Observable trajectories. Length is is ``len(tspan)`` and record names
        follow ``model.observables`` names.
    yobs_view : numpy.ndarray
        An array view (sharing the same data buffer) on ``yobs``. Dimensionality
        is ``(len(tspan), len(model.observables))``.
    integrator : scipy.integrate.ode
        Integrator object.

    Notes
    -----
    The expensive step of generating the code for the right-hand side of the
    model's ODEs is performed during initialization. If you need to integrate
    the same model repeatedly with different parameters then you should build a
    single Solver object and then call its ``run`` method as needed.

    """

    @staticmethod
    def _test_inline():
        """Detect whether scipy.weave.inline is functional."""
        if not hasattr(Solver, '_use_inline'):
            Solver._use_inline = False
            try:
                inline('int i;', force=1)
                Solver._use_inline = True
            except (scipy.weave.build_tools.CompileError,
                    distutils.errors.CompileError, ImportError):
                pass

    def __init__(self, model, tspan, integrator='vode', **integrator_options):

        pysb.bng.generate_equations(model)

        code_eqs = '\n'.join(['ydot[%d] = %s;' %
                              (i, sympy.ccode(model.odes[i]))
                              for i in range(len(model.odes))])
        code_eqs = re.sub(r's(\d+)',
                          lambda m: 'y[%s]' % (int(m.group(1))), code_eqs)
        for e in model.expressions:
            code_eqs = re.sub(r'\b(%s)\b' % e.name,
                              sympy.ccode(e.expand_expr()), code_eqs)
        for i, p in enumerate(model.parameters):
            code_eqs = re.sub(r'\b(%s)\b' % p.name, 'p[%d]' % i, code_eqs)

        Solver._test_inline()
        # If we can't use weave.inline to run the C code, compile it as Python
        # code instead for use with exec. Note: C code with array indexing,
        # basic math operations, and pow() just happens to also be valid
        # Python.  If the equations ever have more complex things in them, this
        # might fail.
        if not Solver._use_inline:
            code_eqs_py = compile(code_eqs, '<%s odes>' % model.name, 'exec')
        else:
            for arr_name in ('ydot', 'y', 'p'):
                macro = arr_name.upper() + '1'
                code_eqs = re.sub(r'\b%s\[(\d+)\]' % arr_name,
                                  '%s(\\1)' % macro, code_eqs)

        def rhs(t, y, p):
            ydot = self.ydot
            # note that the evaluated code sets ydot as a side effect
            if Solver._use_inline:
                inline(code_eqs, ['ydot', 't', 'y', 'p']);
            else:
                exec code_eqs_py in locals()
            return ydot

        # build integrator options list from our defaults and any kwargs passed
        # to this function
        options = {}
        try:
            options.update(default_integrator_options[integrator])
        except KeyError as e:
            pass
        options.update(integrator_options)

        self.model = model
        self.tspan = tspan
        self.y = numpy.ndarray((len(tspan), len(model.species)))
        self.ydot = numpy.ndarray(len(model.species))
        if len(model.observables):
            self.yobs = numpy.ndarray(len(tspan), zip(model.observables.keys(),
                                                      itertools.repeat(float)))
        else:
            self.yobs = numpy.ndarray((len(tspan), 0))
        exprs = model.expressions_dynamic()
        if len(exprs):
            self.yexpr = numpy.ndarray(len(tspan), zip(exprs.keys(),
                                                       itertools.repeat(float)))
        else:
            self.yexpr = numpy.ndarray((len(tspan), 0))
        self.yobs_view = self.yobs.view(float).reshape(len(self.yobs), -1)
        self.integrator = ode(rhs).set_integrator(integrator, **options)


    def run(self, param_values=None, y0=None):
        """Perform an integration.

        Returns nothing; access the Solver object's ``y``, ``yobs``, or
        ``yobs_view`` attributes to retrieve the results.

        Parameters
        ----------
        param_values : vector-like, optional
            Values to use for every parameter in the model. Ordering is
            determined by the order of model.parameters. If not specified,
            parameter values will be taken directly from model.parameters.
        y0 : vector-like, optional
            Values to use for the initial condition of all species. Ordering is
            determined by the order of model.species. If not specified, initial
            conditions will be taken from model.initial_conditions (with initial
            condition parameter values taken from `param_values` if specified).

        """

        if param_values is not None:
            # accept vector of parameter values as an argument
            if len(param_values) != len(self.model.parameters):
                raise Exception("param_values must be the same length as "
                                "model.parameters")
            if not isinstance(param_values, numpy.ndarray):
                param_values = numpy.array(param_values)
        else:
            # create parameter vector from the values in the model
            param_values = numpy.array([p.value for p in self.model.parameters])

        subs = dict((p, param_values[i])
                    for i, p in enumerate(self.model.parameters))
        if y0 is not None:
            # accept vector of species amounts as an argument
            if len(y0) != self.y.shape[1]:
                raise Exception("y0 must be the same length as model.species")
            if not isinstance(y0, numpy.ndarray):
                y0 = numpy.array(y0)
        else:
            y0 = numpy.zeros((self.y.shape[1],))
            for cp, value_obj in self.model.initial_conditions:
                if value_obj in self.model.parameters:
                    pi = self.model.parameters.index(value_obj)
                    value = param_values[pi]
                elif value_obj in self.model.expressions:
                    value = value_obj.expand_expr().evalf(subs=subs)
                else:
                    raise ValueError("Unexpected initial condition value type")
                si = self.model.get_species_index(cp)
                if si is None:
                    raise Exception("Species not found in model: %s" % repr(cp))
                y0[si] = value

        # perform the actual integration
        self.integrator.set_initial_value(y0, self.tspan[0])
        self.integrator.set_f_params(param_values)
        self.y[0] = y0
        i = 1
        while (self.integrator.successful() and
               self.integrator.t < self.tspan[-1]):
            self.y[i] = self.integrator.integrate(self.tspan[i])
            i += 1
        if self.integrator.t < self.tspan[-1]:
            self.y[i:, :] = 'nan'

        for i, obs in enumerate(self.model.observables):
            self.yobs_view[:, i] = \
                (self.y[:, obs.species] * obs.coefficients).sum(1)
        obs_names = self.model.observables.keys()
        obs_dict = dict((k, self.yobs[k]) for k in obs_names)
        for expr in self.model.expressions_dynamic():
            expr_subs = expr.expand_expr().subs(subs)
            func = sympy.lambdify(obs_names, expr_subs, "numpy")
            self.yexpr[expr.name] = func(**obs_dict)


def odesolve(model, tspan, param_values=None, y0=None, integrator='vode',
             **integrator_options):
    """Integrate a model's ODEs over a given timespan.

    This is a simple function-based interface to integrating (a.k.a. solving or
    simulating) a model. If you need to integrate a model repeatedly with
    different parameter values or initial conditions (as in parameter
    estimation), using the Solver class directly will provide much better
    performance.

    Parameters
    ----------
    model : pysb.Model
        Model to integrate.
    tspan : vector-like
        Time values over which to integrate. The first and last values define
        the time range, and the returned trajectories will be sampled at every
        value.
    param_values : vector-like, optional
        Values to use for every parameter in the model. Ordering is determined
        by the order of model.parameters. If not specified, parameter values
        will be taken directly from model.parameters.
    y0 : vector-like, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initial_conditions (with initial
        condition parameter values taken from `param_values` if specified).
    integrator : string, optional
        Name of the integrator to use, taken from the list of integrators known
        to :py:class:`scipy.integrate.ode`.
    integrator_params
        Additional parameters for the integrator.

    Returns
    -------
    yfull : record array
        The trajectories calculated by the integration. The first dimension is
        time and its length is identical to that of `tspan`. The second
        dimension is species/observables and its length is the sum of the
        lengths of model.species and model.observables. The dtype of the array
        specifies field names: '__s0', '__s1', etc. for the species and
        observable names for the observables. See Notes below for further
        explanation and caveats.

    Notes
    -----
    This function was the first implementation of integration support and
    accordingly it has a few warts:

    * It performs expensive code generation every time it is called.

    * The returned array, with its record-style data-type, allows convenient
      selection of individual columns by their field names, but does not permit
      slice ranges or indexing by integers for columns. If you only need access
      to your model's observables this is usually not a problem, but sometimes
      it's more convenient to have a "regular" array. See Examples below for
      code to do this.

    The actual integration code has since been moved to the Solver class and
    split up such that the code generation is only performed on
    initialization. The model may then be integrated repeatedly with different
    parameter values or initial conditions with much better
    performance. Additionally, Solver makes the species trajectories available
    as a simple array and only uses the record array for the observables where
    it makes sense.

    This function now simply serves as a wrapper for creating a Solver object,
    calling its ``run`` method, and building the record array to return.

    Examples
    --------
    Simulate a model and display the results for an observable:

    >>> from pysb.examples.robertson import model
    >>> from numpy import linspace
    >>> numpy.set_printoptions(precision=4)
    >>> yfull = odesolve(model, linspace(0, 40, 10))
    >>> print yfull['A_total']   #doctest: +NORMALIZE_WHITESPACE
    [ 1.      0.899   0.8506  0.8179  0.793   0.7728  0.7557  0.7408  0.7277
    0.7158]

    Obtain a view on a returned record array which uses an atomic data-type and
    integer indexing (note that the view's data buffer is shared with the
    original array so there is no extra memory cost):

    >>> print yfull.shape
    (10,)
    >>> print yfull.dtype   #doctest: +NORMALIZE_WHITESPACE
    [('__s0', '<f8'), ('__s1', '<f8'), ('__s2', '<f8'), ('A_total', '<f8'),
    ('B_total', '<f8'), ('C_total', '<f8')]
    >>> print yfull[0:4, 1:3]   #doctest: +ELLIPSIS
    Traceback (most recent call last):
      ...
    IndexError: too many indices...
    >>> yarray = yfull.view(float).reshape(len(yfull), -1)
    >>> print yarray.shape
    (10, 6)
    >>> print yarray.dtype
    float64
    >>> print yarray[0:4, 1:3]
    [[  0.0000e+00   0.0000e+00]
     [  2.1672e-05   1.0093e-01]
     [  1.6980e-05   1.4943e-01]
     [  1.4502e-05   1.8209e-01]]

    """

    solver = Solver(model, tspan, integrator, **integrator_options)
    solver.run(param_values, y0)

    species_names = ['__s%d' % i for i in range(solver.y.shape[1])]
    yfull_dtype = zip(species_names, itertools.repeat(float))
    if len(solver.yobs.dtype):
        yfull_dtype += solver.yobs.dtype.descr
    yfull = numpy.ndarray(len(solver.y), yfull_dtype)

    yfull_view = yfull.view(float).reshape(len(yfull), -1)
    yfull_view[:, :solver.y.shape[1]] = solver.y
    yfull_view[:, solver.y.shape[1]:] = solver.yobs_view

    return yfull


def setup_module(module):
    """Doctest fixture for nose."""
    # Distutils' temp directory creation code has a more-or-less unsuppressable
    # print to stdout which will break the doctest which triggers it (via
    # scipy.weave.inline). So here we run an end-to-end test of the inlining
    # system to get that print out of the way at a point where it won't matter.
    # As a bonus, the test harness is suppressing writes to stdout at this time
    # anyway so the message is just swallowed silently.
    Solver._test_inline()



"""Translation of the BioNetGen example "Simple" from the BNG wiki.

http://bionetgen.org/index.php/Simple
"""

from pysb import *

Model()


# Physical and geometric constants
Parameter('NA', 6.0e23)      # Avogadro's num
Parameter('f', 0.01)         # scaling factor
Expression('Vo', f * 1e-10)  # L
Expression('V', f * 3e-12)   # L

# Initial concentrations
Parameter('EGF_conc', 2e-9)             # nM
Expression('EGF0', EGF_conc * NA * Vo)  # nM
Expression('EGFR0', f * 1.8e5)          # copy per cell

# Rate constants
Expression('kp1', 9.0e7 / (NA * Vo))  # input /M/sec
Parameter('km1', 0.06)                # /sec


Monomer('EGF', ['R'])
Monomer('EGFR', ['L', 'CR1', 'Y1068'], {'Y1068': ['U', 'P']})


Initial(EGF(R=None), EGF0)
Initial(EGFR(L=None, CR1=None, Y1068='U'), EGFR0)


Rule('egf_binds_egfr', EGF(R=None) + EGFR(L=None) <> EGF(R=1) % EGFR(L=1), kp1, km1)


# Species LR EGF(R!1).EGFR(L!1)
Observable('Lbound', EGF(R=ANY))  # Molecules


if __name__ == '__main__':
    print __doc__, "\n", model
    print "\nNOTE: This model code is designed to be imported and programatically " \
        "manipulated,\nnot executed directly. The above output is merely a " \
        "diagnostic aid."



"""A simple three-species chemical kinetics system known as "Robertson's
example", as presented in:

H. H. Robertson, The solution of a set of reaction rate equations, in Numerical
Analysis: An Introduction, J. Walsh, ed., Academic Press, 1966, pp. 178-182.
"""

# exported from PySB model 'robertson'

import numpy
import scipy.weave, scipy.integrate
import collections
import itertools
import distutils.errors


_use_inline = False
# try to inline a C statement to see if inline is functional
try:
    scipy.weave.inline('int i;', force=1)
    _use_inline = True
except distutils.errors.CompileError:
    pass

Parameter = collections.namedtuple('Parameter', 'name value')
Observable = collections.namedtuple('Observable', 'name species coefficients')
Initial = collections.namedtuple('Initial', 'param_index species_index')


class Model(object):
    
    def __init__(self):
        self.y = None
        self.yobs = None
        self.integrator = scipy.integrate.ode(self.ode_rhs)
        self.integrator.set_integrator('vode', method='bdf',
                                       with_jacobian=True)
        self.y0 = numpy.empty(3)
        self.ydot = numpy.empty(3)
        self.sim_param_values = numpy.empty(6)
        self.parameters = [None] * 6
        self.observables = [None] * 3
        self.initial_conditions = [None] * 3
    
        self.parameters[0] = Parameter('k1', 0.040000000000000001)
        self.parameters[1] = Parameter('k2', 30000000)
        self.parameters[2] = Parameter('k3', 10000)
        self.parameters[3] = Parameter('A_0', 1)
        self.parameters[4] = Parameter('B_0', 0)
        self.parameters[5] = Parameter('C_0', 0)

        self.observables[0] = Observable('A_total', [0], [1])
        self.observables[1] = Observable('B_total', [1], [1])
        self.observables[2] = Observable('C_total', [2], [1])

        self.initial_conditions[0] = Initial(3, 0)
        self.initial_conditions[1] = Initial(4, 1)
        self.initial_conditions[2] = Initial(5, 2)

    if _use_inline:
        
        def ode_rhs(self, t, y, p):
            ydot = self.ydot
            scipy.weave.inline(r'''                
                ydot[0] = -p[0]*y[0] + p[2]*y[1]*y[2];
                ydot[1] = p[0]*y[0] - p[1]*pow(y[1], 2) - p[2]*y[1]*y[2];
                ydot[2] = p[1]*pow(y[1], 2);
                ''', ['ydot', 't', 'y', 'p'])
            return ydot
        
    else:
        
        def ode_rhs(self, t, y, p):
            ydot = self.ydot
            ydot[0] = -p[0]*y[0] + p[2]*y[1]*y[2]
            ydot[1] = p[0]*y[0] - p[1]*pow(y[1], 2) - p[2]*y[1]*y[2]
            ydot[2] = p[1]*pow(y[1], 2)
            return ydot
        
    
    def simulate(self, tspan, param_values=None, view=False):
        if param_values is not None:
            # accept vector of parameter values as an argument
            if len(param_values) != len(self.parameters):
                raise Exception("param_values must have length %d" %
                                len(self.parameters))
            self.sim_param_values[:] = param_values
        else:
            # create parameter vector from the values in the model
            self.sim_param_values[:] = [p.value for p in self.parameters]
        self.y0.fill(0)
        for ic in self.initial_conditions:
            self.y0[ic.species_index] = self.sim_param_values[ic.param_index]
        if self.y is None or len(tspan) != len(self.y):
            self.y = numpy.empty((len(tspan), len(self.y0)))
            if len(self.observables):
                self.yobs = numpy.ndarray(len(tspan),
                                zip((obs.name for obs in self.observables),
                                    itertools.repeat(float)))
            else:
                self.yobs = numpy.ndarray((len(tspan), 0))
            self.yobs_view = self.yobs.view(float).reshape(len(self.yobs),
                                                           -1)
        # perform the actual integration
        self.integrator.set_initial_value(self.y0, tspan[0])
        self.integrator.set_f_params(self.sim_param_values)
        self.y[0] = self.y0
        t = 1
        while self.integrator.successful() and self.integrator.t < tspan[-1]:
            self.y[t] = self.integrator.integrate(tspan[t])
            t += 1
        for i, obs in enumerate(self.observables):
            self.yobs_view[:, i] = \
                (self.y[:, obs.species] * obs.coefficients).sum(1)
        if view:
            y_out = self.y.view()
            yobs_out = self.yobs.view()
            for a in y_out, yobs_out:
                a.flags.writeable = False
        else:
            y_out = self.y.copy()
            yobs_out = self.yobs.copy()
        return (y_out, yobs_out)
    




"""A version of BAX pore assembly where the subunits dimerize and then
tetramerize instead of assembling sequentially (contrast with
bax_pore_sequential.py). Inhibition of pore formation by Mcl-1 is also
implemented.
"""

from pysb import *

Model()

# Each BAX-BAX bond must always involve a t1 site on one monomer and a
# t2 site on the other.
Monomer('BAX', ['t1', 't2', 'inh'])
Annotation(BAX, 'http://identifiers.org/uniprot/Q07812')
Monomer('MCL1', ['b'])
Annotation(MCL1, 'http://identifiers.org/uniprot/Q07820')

# Two lone monomers form a dimer.
Parameter('kdimf', 1e-6)
Parameter('kdimr', 1e-7)
Rule('bax_dim',
     BAX(t1=None, t2=None) + BAX(t1=None, t2=None) <>
     BAX(t1=1, t2=None) % BAX(t1=None, t2=1),
     kdimf, kdimr)

# Two lone dimers form a tetramer, with a higher rate than the dimerization.
Parameter('ktetf', 1e-3)
Parameter('ktetr', 1e-4)
Rule('bax_tet',
     BAX(t1=1, t2=None) % BAX(t1=None, t2=1) + BAX(t1=2, t2=None) % BAX(t1=None, t2=2) <>
     BAX(t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2),
     ktetf, ktetr)

# An inhibitory protein can bind to a BAX subunit at any time.
Parameter('kbaxmcl1f', 1e-5)
Parameter('kbaxmcl1r', 1e-6)
Rule('bax_inh_mcl1',
     BAX(inh=None) + MCL1(b=None) <>
     BAX(inh=1)    % MCL1(b=1),
     kbaxmcl1f, kbaxmcl1r)

# Initial conditions
Parameter('BAX_0', 8e4)
Initial(BAX(t1=None, t2=None, inh=None), BAX_0)
Parameter('MCL1_0', 2e4)
Initial(MCL1(b=None), MCL1_0)
for p in BAX_0, MCL1_0:
    Annotation(p, 'http://identifiers.org/doi/10.1371/journal.pcbi.1002482',
               'isDescribedBy')

# We must fully specify all four BAX-BAX bonds, otherwise the pattern
# is too loose, match a given species multiple times (beyond the
# factor of four expected due to the rotational symmetry of the
# tetramer), resulting in erroneously high values.
Observable('BAX4', BAX(t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2))
# Same all-bonds requirement here.  However since the BAX tetramer is
# considered inhibited when even one subunit has an inhibitor bound,
# we only need to explicitly write inh=ANY on one of the monomer
# patterns.
Observable('BAX4_inh', BAX(inh=ANY, t1=1, t2=3) % BAX(t1=4, t2=1) % BAX(t1=2, t2=4) % BAX(t1=3, t2=2))


if __name__ == '__main__':
    print __doc__, "\n", model
    print "\nNOTE: This model code is designed to be imported and programatically " \
        "manipulated,\nnot executed directly. The above output is merely a " \
        "diagnostic aid. Please see\n" \
        "run_bax_pore.py for example usage."



from pysb import ComponentSet
import pysb.core
import inspect
import numpy
import cStringIO

__all__ = ['alias_model_components', 'rules_using_parameter']

def alias_model_components(model=None):
    """Make all model components visible as symbols in the caller's global namespace"""
    if model is None:
        model = pysb.core.SelfExporter.default_model
    caller_globals = inspect.currentframe().f_back.f_globals
    components = dict((c.name, c) for c in model.all_components())
    caller_globals.update(components)

def rules_using_parameter(model, parameter):
    """Return a ComponentSet of rules in the model which make use of the given parameter"""
    cset = ComponentSet()
    for rule in model.rules:
        if rule.rate_forward is parameter or rule.rate_reverse is parameter:
            cset.add(rule)
    return cset


def synthetic_data(model, tspan, obs_list=None, sigma=0.1):
    #from pysb.integrate import odesolve
    from pysb.integrate import Solver
    solver = Solver(model, tspan)
    solver.run()

    # Sample from a normal distribution with variance sigma and mean 1
    # (randn generates a matrix of random numbers sampled from a normal
    # distribution with mean 0 and variance 1)
    # 
    # Note: This modifies yobs_view (the view on yobs) so that the changes 
    # are reflected in yobs (which is returned by the function). Since a new
    # Solver object is constructed for each function invocation this does not
    # cause problems in this case.
    solver.yobs_view *= ((numpy.random.randn(*solver.yobs_view.shape) * sigma) + 1)
    return solver.yobs

def get_param_num(model, name):
    for i in range(len(model.parameters)):
        if model.parameters[i].name == name:
            print i, model.parameters[i]
            break
    return i


def write_params(model,paramarr, name=None):
    """ write the parameters and values to a csv file
    model: a model object
    name: a string with the name for the file, or None to return the content
    """
    if name is not None:
        fobj = open(name, 'w')
    else:
        fobj = cStringIO.StringIO()
    for i in range(len(model.parameters)):
        fobj.write("%s, %.17g\n"%(model.parameters[i].name, paramarr[i]))
    if name is None:
        return fobj.getvalue()

def update_param_vals(model, newvals):
    """update the values of model parameters with the values from a dict. 
    the keys in the dict must match the parameter names
    """
    update = []
    noupdate = []
    for i in model.parameters:
        if i.name in newvals:
            i.value = newvals[i.name]
            update.append(i.name)
        else:
            noupdate.append(i.name)
    return update, noupdate

def load_params(fname):
    """load the parameter values from a csv file, return them as dict.
    """
    parmsff = {}
    # FIXME: This might fail if a parameter name is larger than 50 characters.
    # FIXME: Maybe do this with the csv module instead?
    temparr = numpy.loadtxt(fname, dtype=([('a','S50'),('b','f8')]), delimiter=',') 
    for i in temparr:
        parmsff[i[0]] = i[1]
    return parmsff



from pysb.bng import generate_network, NoInitialConditionsError
from pysb.core import SelfExporter
import traceback
import os
import importlib

def test_generate_network():
    """Run network generation on all example models"""
    for model in get_example_models():
        yield (check_generate_network, model)

def get_example_models():
    """Generator that yields the model objects for all example models"""
    example_dir = os.path.join(os.path.dirname(__file__), '..', 'examples')
    for filename in os.listdir(example_dir):
        if filename.endswith('.py') and not filename.startswith('run_') \
               and not filename.startswith('__'):
            modelname = filename[:-3]  # strip .py
            package = 'pysb.examples.' + modelname
            module = importlib.import_module(package)
            # Reset do_export to the default in case the model changed it.
            # FIXME the self-export mechanism should be more self-contained so
            # this isn't needed here.
            SelfExporter.do_export = True
            yield module.model

expected_exceptions = {
    'tutorial_b': NoInitialConditionsError,
    'tutorial_c': NoInitialConditionsError,
    }

def check_generate_network(model):
    """Tests that network generation runs without error for the given model"""
    success = False
    try:
        generate_network(model)
        success = True
    except Exception as e:
        # Some example models are deliberately incomplete, so here we will treat
        # any of these "expected" exceptions as a success.
        model_base_name = model.name.rsplit('.', 1)[1]
        exception_class = expected_exceptions.get(model_base_name)
        if exception_class and isinstance(e, exception_class):
            success = True
    assert success, "Network generation failed on model %s:\n-----\n%s" % \
                (model.name, traceback.format_exc())



import pysb

class KappaGenerator(object):

    # Dialect can be either 'complx' or 'kasim' (default)
    def __init__(self, model, dialect='kasim'):
        self.model = model
        self.__content = None
        self.dialect = dialect

    def get_content(self):
        if self.__content == None:
            self.generate_content()
        return self.__content

    def generate_content(self):
        self.__content = ''
        #self.generate_parameters()
        #self.generate_compartments()

        # Agent declarations appear to be required in kasim
        # but prohibited in complx
        if (self.dialect == 'kasim'):
            self.generate_molecule_types() 

        self.generate_reaction_rules()
        self.generate_observables()
        self.generate_species()

    #def generate_parameters(self):
    #    self.__content += "begin parameters\n"
    #    max_length = max(len(p.name) for p in self.model.parameters)
    #    for p in self.model.parameters:
    #        self.__content += ("  %-" + str(max_length) + "s   %e\n") % (p.name, p.value)
    #    self.__content += "end parameters\n\n"

    #def generate_compartments(self):
    #    self.__content += "begin compartments\n"
    #    for c in self.model.compartments:
    #        if c.parent is None:
    #            parent_name = ''
    #        else:
    #            parent_name = c.parent.name
    #        self.__content += ("  %s  %d  %f  %s\n") % (c.name, c.dimension, c.size, parent_name)
    #    self.__content += "end compartments\n\n"        

    def generate_molecule_types(self):
        for m in self.model.monomers:
            site_code = ','.join([format_monomer_site(m, s) for s in m.sites])
            self.__content += "%%agent: %s(%s)\n" % (m.name, site_code)
        self.__content += "\n"

    def generate_reaction_rules(self):
        #self.__content += "begin reaction rules\n"
        #max_length = max(len(r.name) for r in self.model.rules) + 1  # +1 for the colon
        max_length = 0
        for r in self.model.rules:
            label = '\'' + r.name + '\''
            reactants_code = format_reactionpattern(r.reactant_pattern)
            products_code  = format_reactionpattern(r.product_pattern)
            arrow = '->'
            self.__content += ("%s %s %s %s @ %s") % \
                (label, reactants_code, arrow, products_code, float(r.rate_forward.value))
            self.__content += "\n"

            if r.is_reversible:
              label = '\'' + r.name + '_rev' + '\''
              self.__content += ("%s %s %s %s @ %s") % \
                (label, products_code, arrow, reactants_code, float(r.rate_reverse.value))
              self.__content += "\n"

        self.__content += "\n"

    def generate_observables(self):
        if not self.model.observables:
            return
        #self.__content += "begin observables\n"
        for obs in self.model.observables:
            name = '\'' + obs.name + '\''
            observable_code = format_reactionpattern(obs.reaction_pattern)
            self.__content += ("%%obs: %s %s\n") % (name, observable_code)
        self.__content += "\n"
        #self.__content += "end observables\n\n"

    def generate_species(self):
        if not self.model.initial_conditions:
            raise Exception("BNG generator requires initial conditions.")
        species_codes = [format_complexpattern(cp) for cp, param in self.model.initial_conditions]
        #max_length = max(len(code) for code in species_codes)
        max_length = 0
        #self.__content += "begin species\n"
        for i, code in enumerate(species_codes):
            param = self.model.initial_conditions[i][1]
            #self.__content += ("%%init:  %-" + str(max_length) + "s   %s\n") % (code, param.name)
            if (self.dialect == 'kasim'):
                # Switched from %g (float) to %d (int) because kappa didn't like scientific notation
                # for large integers
                self.__content += ("%%init: %d %s\n") % (float(param.value), code)
                #self.__content += ("%%init: %10g %s\n") % (param.value, code)
            else:
                # Switched from %g (float) to %d (int) because kappa didn't like scientific notation
                # for large integers
                self.__content += ("%%init: %10d * %s\n") % (float(param.value), code)
                #self.__content += ("%%init: %10g * %s\n") % (param.value, code)

        self.__content += "\n"


def format_monomer_site(monomer, site):
    ret = site
    if monomer.site_states.has_key(site):
        for state in monomer.site_states[site]:
            ret += '~' + state
    return ret

def format_reactionpattern(rp):
    return ','.join([format_complexpattern(cp) for cp in rp.complex_patterns])

def format_complexpattern(cp):
    ret = ','.join([format_monomerpattern(mp) for mp in cp.monomer_patterns])
    if cp.compartment is not None:
        ret = '@%s:%s' % (cp.compartment.name, ret)
    return ret

def format_monomerpattern(mp):
    # sort sites in the same order given in the original Monomer
    site_conditions = sorted(mp.site_conditions.items(),
                             key=lambda x: mp.monomer.sites.index(x[0]))
    site_pattern_code = ','.join([format_site_condition(site, state) for (site, state) in site_conditions])
    ret = '%s(%s)' % (mp.monomer.name, site_pattern_code)
    if mp.compartment is not None:
        ret = '%s@%s' % (ret, mp.compartment.name)
    return ret

def format_site_condition(site, state):
    if state == None:
        state_code = ''
    elif type(state) == int:
        state_code = '!' + str(state)
    elif type(state) == list:
        raise Exception("Kappa generator does not support multiple bonds to a single site.")
    elif type(state) == str:
        state_code = '~' + state
    elif type(state) == tuple:
        if state[1] == pysb.WILD:
            state = (state[0], '?')
        state_code = '~%s!%s' % state
    elif state == pysb.ANY:
        state_code = '!_'
    else:
        raise Exception("Kappa generator has encountered an unknown element in a rule pattern site condition.")
    return '%s%s' % (site, state_code)



#!/usr/bin/env python
"""
Usage: ``python -m pysb.tools.render_species mymodel.py > mymodel.dot``

Renders the species from a model into the "dot" graph format which can be
visualized with Graphviz.

To create a PDF from the .dot file, use the Graphviz tools in the following
command pipeline::

    ccomps -x mymodel.dot | dot | gvpack -m0 | neato -n2 -T pdf -o mymodel.pdf

You can also change the "dot" command to "circo" or "sfdp" for a different type
of layout. Note that you can pipe the output of render_species straight into a
Graphviz command pipeline without creating an intermediate .dot file, which is
especially helpful if you are making continuous changes to the model and need to
visualize your changes repeatedly::

    python -m pysb.tools.render_species mymodel.py | ccomps -x | dot |
      gvpack -m0 | neato -n2 -T pdf -o mymodel.pdf

Note that some PDF viewers will auto-reload a changed PDF, so you may not even
need to manually reopen it every time you rerun the tool.
"""

import sys
import os
import re
import pygraphviz
import pysb.bng

def run(model):
    """
    Render the species from a model into the "dot" graph format.

    Parameters
    ----------
    model : pysb.core.Model
        The model to render.

    Returns
    -------
    string
        The dot format output.
    """

    pysb.bng.generate_equations(model)
    return render_species_as_dot(model.species, model.name)


def render_species_as_dot(species_list, graph_name=""):
    graph = pygraphviz.AGraph(name="%s species" % graph_name, rankdir="LR",
                              fontname='Arial')
    graph.edge_attr.update(fontname='Arial', fontsize=8)
    for si, cp in enumerate(species_list):
        sgraph_name = 'cluster_s%d' % si
        cp_label = re.sub(r'% ', '%<br align="left"/>',
                          str(cp)) + '<br align="left"/>'
        sgraph_label = '<<font point-size="10" color="blue">s%d</font>' \
                       '<br align="left"/>' \
                       '<font face="Consolas" point-size="6">%s</font>>' % \
                       (si, cp_label)
        sgraph = graph.add_subgraph(name=sgraph_name, label=sgraph_label,
                                    color="gray75", sortv=sgraph_name)
        bonds = {}
        for mi, mp in enumerate(cp.monomer_patterns):
            monomer_node = '%s_%d' % (sgraph_name, mi)
            monomer_label = '<<table border="0" cellborder="1" cellspacing="0">'
            monomer_label += '<tr><td bgcolor="#a0ffa0"><b>%s</b></td></tr>' % \
                             mp.monomer.name
            for site in mp.monomer.sites:
                site_state = None
                cond = mp.site_conditions[site]
                if isinstance(cond, str):
                    site_state = cond
                elif isinstance(cond, tuple):
                    site_state = cond[0]
                site_label = site
                if site_state is not None:
                    site_label += '=<font color="purple">%s</font>' % site_state
                monomer_label += '<tr><td port="%s">%s</td></tr>' % \
                                 (site, site_label)
            for site, value in mp.site_conditions.items():
                site_bonds = []  # list of bond numbers
                if isinstance(value, int):
                    site_bonds.append(value)
                elif isinstance(value, tuple):
                    site_bonds.append(value[1])
                elif isinstance(value, list):
                    site_bonds += value
                for b in site_bonds:
                    bonds.setdefault(b, []).append((monomer_node, site))
            monomer_label += '</table>>'
            sgraph.add_node(monomer_node,
                            label=monomer_label, shape="none", fontname="Arial",
                            fontsize=8)
        for bi, sites in bonds.items():
            node_names, port_names = zip(*sites)
            sgraph.add_edge(node_names, tailport=port_names[0],
                            headport=port_names[1], label=str(bi))
    return graph.string()

usage = __doc__
usage = usage[1:]  # strip leading newline

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        print usage,
        exit()
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble (use the imp package and import
        # as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print run(model)



"""
Wrapper functions for running the Kappa programs *Kasim* and *complx*.

In general only the following three functions will be needed for typical use:

    * :py:func:`run_simulation`
    * :py:func:`influence_map`
    * :py:func:`contact_map`

The other functions are used internally and manage the execution of the Kappa
software and the parsing of the data into a Numpy format.
"""

__author__ = "johnbachman"

import pysb
from pysb.generator.kappa import KappaGenerator
import os
import subprocess
import random
import re
import sympy
import numpy as np

def run_simulation(model, **kwargs):
    """Runs the given model using KaSim and returns the parsed results.

    Parameters
    ----------
    **kwargs : List of keyword arguments
        All keyword arguments specifying conditions for the simulation are
        passed to the function :py:func:`run_kasim` (see documentation
        associated with that function for more information).

    Returns
    -------
    numpy.ndarray
        Returns the kasim simulation data as a Numpy ndarray. Data is accessed
        using the syntax::

            results[index_name]

        The index 'time' gives the data for the time coordinates of the
        simulation. Data for the observables can be accessed by indexing the
        array with the names of the observables.
    """

    outs = run_kasim(model, **kwargs)
    return parse_kasim_outfile(outs['out'])

def influence_map(model, do_open=False, **kwargs):
    """Generates the influence map.

    Runs KaSim with no events or points and sets the dump_influence_map
    argument to True.

    Parameters
    ----------
    model : pysb.core.Model
        The model for generating the influence map.
    do_open : boolean
        If do_open is set to True, then calls the :py:func:`open_file` method
        to display the influence map using the default program for opening .gv
        files (e.g., GraphViz).
    **kwargs : other keyword arguments
        Any other keyword arguments are passed to the function
        :py:func:`run_kasim`.

    Returns
    -------
    string
        Returns the name of the .gv (GraphViz) file where the influence map
        has been stored. This can subsequently be used to build a networkx or
        PyGraphViz graph.
    """

    kasim_dict = run_kasim(model, time=0, points=0, dump_influence_map=True,
                           **kwargs)
    im_filename = kasim_dict['im']

    if do_open:
        open_file(im_filename)

    return im_filename

def contact_map(model, output_dir='.', base_filename=None, do_open=False, 
                **kwargs):
    """Runs complx with arguments for generating the contact map.

    Parameters
    ----------
    model : pysb.core.Model
        The model for generating the contact map.
    output_dir : string
        The subdirectory in which to generate the Kappa (.ka) file for the
        model and all output files produced by complx. Default value is '.'
        Note that only relative paths can be specified; paths are relative
        to the directory where the current Python instance is running.
        If the specified directory does not exist, an Exception is thrown.
    base_filename : string
        The base filename to be used for generation of the Kappa (.ka) file and
        all output files produced by complx. Defaults to a string of the form::

            '%s_%d_%d_temp' % (model.name, os.getpid(), random.randint(0,10000))

        The contact map filenames append '_cm.jpg' and '_cm.dot' to this base
        filename; the reachable complexes filename appends '_rch.dot'.
    do_open : boolean
        If do_open is set to True, then calls the :py:func:`open_file` method
        to display the contact map using the default program for opening .jpg
        files.
    **kwargs : other keyword arguments
        Any other keyword arguments are passed through to the function
        :py:func:`run_complx`.

    """

    gen = KappaGenerator(model, dialect='complx')

    if not base_filename:
        base_filename = '%s/%s_%d_%d_temp' % (output_dir,
                         model.name, os.getpid(), random.randint(0, 10000))

    kappa_filename = base_filename + '.ka'
    jpg_filename = base_filename + '_cm.jpg'
    dot_filename = base_filename + '_cm.dot'
    reachables_filename = base_filename + '_rch.dot'

    args = ['--output-high-res-contact-map-jpg', jpg_filename,
            '--output-high-res-contact-map-dot', dot_filename,
            '--output-reachable-complexes', reachables_filename]
    run_complx(gen, kappa_filename, args, **kwargs)

    if do_open:
        open_file(jpg_filename)


### "PRIVATE" Functions ###############################################

def run_complx(gen, kappa_filename, args):
    """Generalized method for passing arguments to the complx executable.

    Parameters
    ----------
    gen : :py:class:`pysb.generator.KappaGenerator`
        A KappaGenerator object that is used to produce the Kappa content
        for writing to a file.
    kappa_filename : string
        The name of the file to write the generated Kappa to.
    args : list of strings
        List of command line arguments to pass to complx, with one entry for
        each argument, for example::

            ['--output-high-res-contact-map-jpg', jpg_filename]
    """

    try:
        kappa_file = open(kappa_filename, 'w')
        kappa_file.write(gen.get_content())
        kappa_file.close()
        cmd = 'complx ' + ' '.join(args) + ' ' + kappa_filename
        print "Command: " + cmd
        p = subprocess.Popen(['complx'] + args + [kappa_filename],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        #p.communicate()
        p.wait()

        if p.returncode:
            raise Exception(p.stderr.read())

    except Exception as e:
        raise Exception("problem running complx: " + str(e))


def run_kasim(model, time=10000, points=200, output_dir='.', cleanup=False,
              base_filename=None, dump_influence_map=False,
              perturbation=None):
    """Run KaSim on the given model with the provided arguments.

    Parameters
    ----------
    model : pysb.core.Model
        The model to simulate/analyze using KaSim.
    time : number
        The amount of time (in arbitrary units) to run a simulation.
        Identical to the -t argument when using KaSim at the command line.
        Default value is 10000. If set to 0, no simulation will be run, but
        the influence map will be generated (if dump_influence_map is set to
        True).
    points : integer
        The number of data points to collect for plotting.
        Identical to the -p argument when using KaSim at the command line.
        Default value is 200.
    output_dir : string
        The subdirectory in which to generate the Kappa (.ka) file for the
        model and all output files produced by KaSim. Default value is '.'
        Note that only relative paths can be specified; paths are relative
        to the directory where the current Python instance is running.
        If the specified directory does not exist, an Exception is thrown.
    cleanup : boolean
        Specifies whether output files produced by KaSim should be deleted
        after execution is completed. Default value is False.
    base_filename : The base filename to be used for generation of the Kappa
        (.ka) file and all output files produced by KaSim. Defaults to a
        string of the form::

            '%s_%d_%d_temp' % (model.name, program id, random.randint(0,10000))

        The influence map filename appends '_im.dot' to this base filename; the
        flux map filename appends '_fm.dot'; and the simulation output file
        appends '.out'
    dump_influence_map : boolean
        Specifies whether or not to produce the influence map. Default value
        is False.
    perturbation : string or None
        Optional perturbation language syntax to be appended to the Kappa file.
        See KaSim manual for more details. Default value is None (no
        perturbation).

    Returns
    -------
    A dict with three entries giving the filenames for the files produced:

        * output_dict['out'] gives the .out filename
        * output_dict['im'] gives the influence map filename
        * output_dict['fm'] gives the flux map filename
    """

    gen = KappaGenerator(model)

    if not base_filename:
        base_filename = '%s/%s_%d_%d_temp' % (output_dir,
                        model.name, os.getpid(), random.randint(0, 10000))

    kappa_filename = base_filename + '.ka'
    im_filename = base_filename + '_im.dot'
    fm_filename = base_filename + '_fm.dot'
    out_filename = base_filename + '.out'

    args = ['-i', kappa_filename, '-t', str(time), '-p', str(points),
            '-o', out_filename]

    try:
        kappa_file = open(kappa_filename, 'w')

        # Generate the Kappa model code from the PySB model and write it to
        # the Kappa file:
        kappa_file.write(gen.get_content())

        # If desired, add instructions to the kappa file to generate the
        # influence map:
        if dump_influence_map:
            kappa_file.write('%def: "dumpInfluenceMap" "true"\n')
            kappa_file.write('%%def: "influenceMapFileName" "%s"\n\n' %
                             im_filename)

        # If any perturbation language code has been passed in, add it to the
        # Kappa file:
        if perturbation:
            kappa_file.write('\n%s\n' % perturbation)

        kappa_file.close()

        print "Running kasim"
        p = subprocess.Popen(['KaSim'] + args)
                           #stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        p.communicate()

        if p.returncode:
            raise Exception(p.stdout.read())

    except Exception as e:
        raise Exception("Problem running KaSim: " + str(e))

    finally:
        if cleanup:
            for filename in [kappa_filename, im_filename,
                            fm_filename, out_filename]:
                if os.access(filename, os.F_OK):
                    os.unlink(filename)

    output_dict = {'out':out_filename, 'im':im_filename, 'fm':'flux.dot'}
    return output_dict


def parse_kasim_outfile(out_filename):
    """
    Parses the KaSim .out file into a Numpy ndarray.

    Parameters
    ----------
    out_filename : string
        String specifying the location of the .out filename produced by KaSim.

    Returns
    -------
    numpy.ndarray
        Returns the KaSim simulation data as a Numpy ndarray. Data is accessed
        using the syntax::

            results[index_name]

        The index 'time' gives the data for the time coordinates of the
        simulation. Data for the observables can be accessed by indexing the
        array with the names of the observables.
    """

    try:
        out_file = open(out_filename, 'r')

        line = out_file.readline().strip() # Get the first line
        out_file.close()
        line = line[2:]  # strip off opening '# '
        raw_names = re.split(' ', line)
        column_names = []

        # Get rid of the quotes surrounding the observable names
        for raw_name in raw_names:
            mo = re.match("'(.*)'", raw_name)
            if (mo): column_names.append(mo.group(1))
            else: column_names.append(raw_name)

        # Create the dtype argument for the numpy record array
        dt = zip(column_names, ('float',)*len(column_names))

        # Load the output file as a numpy record array, skip the name row
        arr = np.loadtxt(out_filename, dtype=dt, skiprows=1)

    except Exception as e:
        raise Exception("problem parsing KaSim outfile: " + str(e))

    return arr


def open_file(filename):
    """Utility function for opening files for display on Mac OS X.

    Uses the 'open' command to open the given file using the default program
    associated with the file's filetype. Ultimately this should be rewritten to
    auto-detect the operating system and use the appropriate system call.
    """

    try:
        p = subprocess.Popen(['open'] + [filename],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        #p.communicate()
        p.wait()
        if p.returncode:
            raise Exception(p.stderr.read())
    except Exception as e:
        raise Exception("Problem opening file: ", e)



#!/usr/bin/env python

from distutils.core import setup
import distutils.cmd
import sys, os, subprocess, traceback, re

def main():

    setup(name='pysb',
          version=get_version(),
          description='Python Systems Biology modeling framework',
          long_description='PySB (pronounced "Pie Ess Bee") is a framework ' + \
              'for building rule-based mathematical models of biochemical ' + \
              'systems. It works nicely with scientific Python libraries ' + \
              'such as NumPy, SciPy and SymPy for model simulation and ' + \
              'analysis.',
          author='Jeremy Muhlich',
          author_email='jmuhlich@bitflood.org',
          url='http://pysb.org/',
          packages=['pysb', 'pysb.generator', 'pysb.tools', 'pysb.examples',
                    'pysb.export', 'pysb.testing', 'pysb.tests'],
          scripts=['scripts/pysb_export'],
          requires=['numpy', 'scipy', 'sympy'],
          keywords=['systems', 'biology', 'model', 'rules'],
          classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Mathematics',
            ],
          )

def get_version():
    rv_filename = 'RELEASE-VERSION'
    base_path = os.path.dirname(__file__)
    rv_path = os.path.join(base_path, rv_filename)
    try:
        version = get_git_version()
        version_file = open(rv_path, 'w')
        version_file.write(version)
        version_file.close() # ensure sdist build process sees new contents
    except Exception:
        try:
            version_file = open(rv_path, 'r')
            version = version_file.read()
        except IOError as e:
            if e.errno == 2 and e.filename == rv_path:
                sys.stderr.write(
"""This does not appear to be a git repository, and the file %s is not
present. In order to build or install PySB, please either download a
distribution from http://pypi.python.org/pypi/pysb or clone the git repository
at https://github.com/pysb/pysb.git\n""" % rv_filename)
                sys.exit(1)
    return version

class GitError(Exception):
    pass

def get_git_version():
    """Get a nice version number from git-describe"""

    # ensure that we are working in a pysb git repo
    setup_path = os.path.abspath(os.path.dirname(__file__))
    if not os.path.exists(os.path.join(setup_path, '.git')):
        raise Exception("setup.py is not in the root of a git repository; "
                        "aborting")
    os.chdir(setup_path)

    # run git describe
    gitcmd = ['git', 'describe', '--always']
    try:
        gitproc = subprocess.Popen(gitcmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        retcode = gitproc.wait()
        if retcode:
            raise GitError(gitproc.stderr.read())
        version = gitproc.stdout.next().strip()
        version = re.sub(r'^v', '', version)
        return version
    except (OSError, GitError) as e:
        raise Exception("Error running 'git describe' to determine version:\n\n" +
                        "command\n=====\n" + " ".join(gitcmd) + "\n\n" +
                        "error\n====\n" + str(e) + "\n")

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from pysb.integrate import odesolve
from pylab import linspace, plot, legend, show

from kinase_cascade import model


tspan = linspace(0, 1200)
print "Simulating..."
yfull = odesolve(model, tspan)
plot(tspan, yfull['ppMEK'], label='ppMEK')
plot(tspan, yfull['ppERK'], label='ppERK')
legend(loc='upper left')
show()



"""Demonstration of the move_connected keyword for Rules
"""

from pysb import *

Model()

Monomer('A', ['b'])
Monomer('B', ['a'])

# One main 3d compartment and two 2d membranes inside it
Parameter('Vmain', 1)
Parameter('Vx', 1)
Parameter('Vy', 1)
Compartment('Main', None, 3, Vmain)
Compartment('X', Main, 2, Vx)
Compartment('Y', Main, 2, Vy)

# A and B, both embedded in membrane X, bind reversibly
Parameter('Kab_f', 1)
Parameter('Kab_r', 1)
Rule('Ax_bind_Bx', A(b=None) ** X + B(a=None) ** X <> A(b=1) ** X % B(a=1) ** X,
     Kab_f, Kab_r)

# The A:B complex is transported back and forth from X to Y
Parameter('Ktrans_f', 1)
Parameter('Ktrans_r', 1)
# move_connected is required or B will be "left behind" and BNG will complain
# (change move_connected to False and run pysb.tools.export_bng_net on this file
# and watch for the WARNING line in the output log)
Rule('ABx_trans_y', A(b=ANY) ** X <> A(b=ANY) ** Y,
     Ktrans_f, Ktrans_r, move_connected=True)

Parameter('ABx_0', 1)
Initial(A(b=1)**X % B(a=1)**X, ABx_0)

if __name__ == '__main__':
    print __doc__, "\n", model
    print "\nNOTE: This model code is designed to be imported and programatically " \
        "manipulated,\nnot executed directly. The above output is merely a " \
        "diagnostic aid."



from pysb.testing import *
from pysb.core import *
from functools import partial

def test_component_names_valid():
    for name in 'a', 'B', 'AbC', 'dEf', '_', '_7', '__a01b__999x_x___':
        c = Component(name, _export=False)
        eq_(c.name, name)

def test_component_names_invalid():
    for name in 'a!', '!B', 'A!bC~`\\', '_!', '_!7', '__a01b  999x_x___!':
        assert_raises(InvalidComponentNameError, Component, name, _export=False)

def test_monomer():
    sites = ['x', 'y', 'z']
    states = {'y': ['foo', 'bar', 'baz'], 'x': ['e']}
    m = Monomer('A', sites, states, _export=False)
    assert_equal(m.sites, sites)
    assert_equal(m.site_states, states)
    assert_equal(type(m()), MonomerPattern)

    assert_raises(ValueError, Monomer, 'A', 'x', _export=False)
    assert_raises(Exception, Monomer, 'A', 'x', 'x', _export=False)
    assert_raises(Exception, Monomer, 'A', ['x'], {'y': ['a']}, _export=False)
    assert_raises(Exception, Monomer, 'A', ['x'], {'x': [1]}, _export=False)

@with_model
def test_monomer_model():
    Monomer('A')
    ok_(A in model.monomers)
    ok_(A in model.all_components())
    ok_(A not in model.all_components() - model.monomers)


@with_model
def test_initial():
    Monomer('A', ['s'])
    Parameter('A_0')
    Initial(A(s=None), A_0)
    assert_raises_iice = partial(assert_raises, InvalidInitialConditionError,
                                 Initial)
    assert_raises_iice('not a complexpattern', A_0)
    assert_raises_iice(A(), A_0)
    assert_raises_iice(A(s=None), A_0)
    assert_raises_iice(MatchOnce(A(s=None)), A_0)

@with_model
def test_model_pickle():
    import pickle
    A = Monomer('A', _export=False)
    B = Monomer('B', ['x', 'y'], {'x': ['e', 'f']}, _export=False)
    k = Parameter('k', 1.0, _export=False)
    r = Rule('r', A() + B(x='e', y=WILD) >> A() % B(x='f', y=None), k,
             _export=False)
    o = Observable('o', A() % B(), _export=False)
    e = Expression('e', k * o, _export=False)
    c = Compartment('c', None, 3, k, _export=False)
    for comp in [A, B, k, r, o, e, c]:
        model.add_component(comp)
    model.add_component(c)
    Initial(A() ** c, k)
    assert_equal(len(model.all_components()), 7)
    model2 = pickle.loads(pickle.dumps(model))
    check_model_against_component_list(model, model2.all_components())

@with_model
def test_monomer_as_reaction_pattern():
    A = Monomer('A', _export=False)
    as_reaction_pattern(A)

@with_model
def test_monomer_as_complex_pattern():
    A = Monomer('A', _export=False)
    as_complex_pattern(A)

@with_model
def test_observable_constructor_with_monomer():
    A = Monomer('A', _export=False)
    o = Observable('o', A, _export=False)

@with_model
def test_compartment_initial_error():
    Monomer('A', ['s'])
    Parameter('A_0', 2.0)
    c1 = Compartment("C1")
    c2 = Compartment("C2")
    Initial(A(s=None)**c1, A_0)
    Initial(A(s=None)**c2, A_0)

@with_model
def test_monomer_pattern_add_to_none():
    """Ensure that MonomerPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    ok_(isinstance(A() + None, ReactionPattern),
        'A() + None did not return a ReactionPattern.')

@with_model
def test_complex_pattern_add_to_none():
    """Ensure that ComplexPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    ok_(isinstance(A(s=1) % A(s=1) + None, ReactionPattern),
        'A(s=1) % A(s=1) + None did not return a ReactionPattern.')

@with_model
def test_reaction_pattern_add_to_none():
    """Ensure that ReactionPattern + None returns a ReactionPattern."""
    Monomer('A', ['s'])
    cp = A(s=1) % A(s=1)
    rp = cp + cp
    ok_(isinstance(rp + None, ReactionPattern),
        'ReactionPattern + None did not return a ReactionPattern.')



#!/usr/bin/env python

import pysb
import pysb.bng
import sympy
import re
import sys
import os
from StringIO import StringIO


def run(model):
    output = StringIO()
    pysb.bng.generate_equations(model)

    ## PARAMETERS
    param_list = '\n'.join(['param(%d) = %s; %% %s' % (i+1, p.value, p.name) for i, p in
       enumerate(model.parameters)])

    param_list_file = open('%s_nominal_values.m' % model.name, 'w')
    param_list_file.write('function param = %s_nominal_values()\n\n' % model.name)
    param_list_file.write(param_list)
    param_list_file.write('\n\nend\n')
    param_list_file.close()

    ## ODES
    odes = 'function out = %s_odes(t, input, param)\n\n' % model.name

    species_list = '\n'.join(['%% input(%d) = %s;' % (i+1, s) for i, s in
        enumerate(model.species)])
    odes += species_list + "\n\n"

    odes += '\n'.join(['%% %s\nout(%d,1) = %s;' % (model.species[i], i+1, sympy.ccode(model.odes[i]))
                       for i in range(len(model.odes))])
    odes = re.sub(r's(\d+)', lambda m: 'input(%s)' % (int(m.group(1))+1), odes)
    odes = re.sub(r'pow\(', 'power(', odes)
    odes += '\n\nend\n'

    # Substitute the parameter references into the ODE expressions
    for i, p in enumerate(model.parameters):
        odes = re.sub(r'\b(%s)\b' % p.name, 'param(%d)' % (i+1), odes)

    ode_file = open('%s_odes.m' % model.name, 'w')
    ode_file.write(odes)
    ode_file.close()

    ## OBSERVABLES
    ## -- hoda
    # function [ode_observables, kd_values, kd_index, ic_index, dividing_factor] = %s_observables()
    # ode_observables is a cell array, with three entries for each index: first is a list of species
    # second is a matched list of coefficients, and the third is a normalization factor (usually an
    # condition) or 1 if not needed
    #
    # kd_values is a cell array where each entry is a tuple of the relevant initial conditions
    # kd_index is a matched cell array with each entry a tuple for the index of the relevant parameters to
    # be perturbed in their initial conditions
    # ic_index=cell array with a list that is mapped to the ODEs (i.e., the init_conds) file
    observables = "function [ode_observables, kd_values, kd_index, "
    observables += "ic_index, dividing_factor] = %s_observables()\n\n" % model.name

    for i, obs in enumerate(model.observables):
        obs_spec_list = "[" + ' '.join(str(s+1) for s in obs.species) + "]"
        obs_coeff_list = "[" + ' '.join(str(c) for c in obs.coefficients) + "]"

        observables += "ode_observables{%d, 1} = %s; %% %s\n" % (i+1, obs_spec_list, str(obs))
        observables += "ode_observables{%d, 2} = %s;\n" % (i+1, obs_coeff_list)
        observables += "ode_observables{%d, 3} = 1;\n" % (i+1)
    observables += "end\n"

    obs_file = open('%s_observables.m' % model.name, 'w')
    obs_file.write(observables)
    obs_file.close()

    ## PRIOR FLAGS
    # function prior_flag=%s_prior_flags()
    # returns array: if the param is an initial condition, set it to 0 in the array
    # otherwise: returns 1 (future implementation to return a value for
    # 1st forsrd, 1st, reverse, 2nd order forward, (2nd reverse), catalytic (1-5)
    # for now default to 1
    ic_names = [ic[1].name for ic in model.initial_conditions]

    prior_flags = "function prior_flag=%s_prior_flags()\n\n" % model.name
    for i, p in enumerate(model.parameters):    
        if (p.name in ic_names):
            prior_flags += "prior_flag(%d) = 0; %% %s\n" % (i+1, p.name)
        else:
            prior_flags += "prior_flag(%d) = 1; %% %s\n" % (i+1, p.name)
    prior_flags += "\nend\n" 

    prior_flags_file = open('%s_prior_flags.m' % model.name, 'w')
    prior_flags_file.write(prior_flags)
    prior_flags_file.close()
 
    ## INITIAL CONDITIONS
    # function conc = %s_init_conds()
    # returns an array that matches the num of ODEs (mostly zeros)
    init_conds = "function conc = %s_init_conds()\n\n" % model.name
    init_conds += "conc = zeros(%d, 1);\n" % len(model.species)
    for ic in model.initial_conditions:
        init_conds += 'conc(%d) = %g; %% %s; %s\n' % (model.get_species_index(ic[0])+1, ic[1].value,
                                                  ic[1].name, ic[0])
    init_conds += "\nend\n\n"

    init_conds_file = open('%s_init_conds.m' % model.name, 'w')
    init_conds_file.write(init_conds)
    init_conds_file.close()

    #output.write(param_list + "\n\n")
    #output.write(c_code_species + "\n\n")
    #output.write(odes + "\n\n")
    #output.write("end\n")
    #return output.getvalue()

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        raise Exception("You must specify the filename of a model script")
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble
        # (use the imp package and import as some safe name?)
        model_module = __import__(model_name)
    except StandardError as e:
        print "Error in model script:\n"
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    # print run(model)
    run(model)



